"""
bass.layout
-----
Objects and functions related to the layout of the rendered pages.
Chameleon is the primary template engine. Other template engines can be added.
"""

from . import setting
from .common import logger
from os import listdir
from os.path import join, splitext, isfile
import sys

# By default, there is one template factory: chameleon.PageTemplateFile. This is associated
# with the file extensions '.xml' and '.pt'. Other template factories can be defined,
# provided they implement the following interface:
#   - filename -> template: template = template_factory(filename)
#   - node     -> string: template.render(this=node) returns HTML page for node 'this'
#                 (condition: node.skin should be equal to filename without extension)
# Template factories are stored in a dictionary template_factory, with the extension as key.

template_factory = {}

def add_template_type(extension, factory):
    """
    add template factory for given extension
    :param extension: file extension (.foo)
    :param factory:   template factory (callable object)
    """
    if extension in template_factory:
        logger.debug(f'Cannot redefine template type {extension}')
    else:
        logger.debug(f'Define new template type {extension}')
        template_factory[extension] = factory

def copy_template_type(from_extension, to_extension):
    """copy existing template factory to another extension"""
    if to_extension in template_factory:
        logger.debug(f'Cannot redefine template type {to_extension}')
    elif from_extension in template_factory:
        logger.debug(f'Template type {to_extension} copied from {from_extension}')
        template_factory[to_extension] = template_factory[from_extension]
    else:
        logger.debug(f'No template type {from_extension}')

try:
    from chameleon import PageTemplateFile
    add_template_type('.xml', PageTemplateFile)
    copy_template_type('.xml', '.pt')
except ImportError:
    logger.critical('Chameleon template engine not available')
    sys.exit(1)

def read_templates():
    """Read templates from layout directory. This function should be called
    just before rendering the site tree and after the extensions have been imported."""
    template = {}
    template_types = list(template_factory.keys())
    logger.debug('Scanning for templates in {}'.format(setting.layout))
    logger.debug('Template types: {}'.format(' '.join(template_types)))
    for filename in listdir(setting.layout):
        (name, extension) = splitext(filename)
        file_path = join(setting.layout, filename)
        # all files (not symbolic links) with a recognized extension are possible templates
        if isfile(file_path) and extension in template_types: # other files are ignored
            try:
                template[name] = template_factory[extension](file_path)
            except Exception as e:
                logger.debug(f'Error in template for {name} in file {filename}')
                logger.debug(str(e))
    if 'default' in template:
        setting.template = template
    else:
        logger.critical('There is no default template')
        sys.exit(1)



"""
bass.common
-----
Basic objects and functions shared by other modules.

Logging
-------
Logging levels are DEBUG, INFO, WARNING, ERROR, CRITICAL.
Bass uses logging level INFO by default, and DEBUG if called with --debug.

Basic (no webob, no waitress):

Webob available:

Webob and Waitress available:
Waitress uses a logger object named 'waitress', and sets the logging level to WARNING.
"""

import logging
# configure logging
logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
try:
    import waitress
    logger = logging.getLogger('waitress')
except ImportError:
    logger = logging.getLogger('bass')

from yaml import load
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

def read_file(filename):
    """read entire file, return content as one string"""
    with open(filename, 'rU') as f:
        try:
            text = ''.join(f.readlines())
        except UnicodeError:
            text = ''
            logger.debug(f'Unicode error in file {filename}')
    return text

def write_file(text, filename):
    """write text to file"""
    with open(filename, 'w') as f:
        f.write(text)

def read_yaml_file(path):
    """read file, return YAML content as dictionary"""
    with open(path, 'r') as f:
        result = load(f, Loader=Loader)
    return result

def read_yaml_string(string):
    """read string, return YAML content as dictionary"""
    result = load(string, Loader=Loader)
    return result



"""
bass.site
-----
Objects and functions related to site structure and site generation.
"""

import shutil, sys, yaml
from . import setting
from .common import write_file
from .config import config_default, read_config
from .event import event_handler
from .layout import read_templates
from .common import logger
from .tree import Folder, Page, Asset
from fnmatch import fnmatch
from importlib import import_module
from os import scandir, mkdir, unlink, walk
from os.path import isdir, isfile, islink, join, relpath, splitext, split

def create_project():
    """create new project directory, with default configuration"""
    logger.info('Creating project')
    if len(list(scandir())) == 0:
        write_file(yaml.dump(config_default, default_flow_style=False), 'config')
        mkdir(config_default['input'])
        mkdir(config_default['output'])
        mkdir(config_default['layout'])
    else:
        logger.warning('Current directory not empty')
        sys.exit()

def build_site():
    """build site in project directory"""
    read_config()
    verify_project()
    read_extension()
    logger.info('Building site tree')
    root = generate_tree()
    prepare_output()
    read_templates()
    logger.info('Rendering site tree')
    root.render()

def rebuild_site():
    """rebuild site in project directory"""
    logger.info('Building modified site tree')
    root = generate_tree()
    prepare_output()
    read_templates()
    logger.info('Rendering modified site tree')
    root.render()

def verify_project():
    """verify existence of directories specified in configuration"""
    if not (isdir(setting.input) and isdir(setting.output) and isdir(setting.layout)):
        logger.critical('Directories missing in project')
        sys.exit(1)

def read_extension():
    """read extension(s) from package specified in configuration file"""
    if setting.extension and isdir(join(setting.project, setting.extension)):
        try:
            logger.debug(f'Adding project directory {setting.project} to Python path')
            sys.path.insert(0, setting.project)
            logger.debug(f'Importing package {setting.extension}')
            module = import_module(setting.extension)
        except ImportError:
            logger.debug(f'Extension directory {setting.extension} is not a Python package')

def prepare_output():
    """clean output directory before rendering site tree"""
    logger.debug(f'Clean output directory {setting.output}')
    for name in [n for n in scandir(setting.output) if n.name != '.']:
        path = join(setting.output, name)
        if isfile(path):
            unlink(path)
        else:
            shutil.rmtree(path)

def ignore_entry(name_rel, dirname):
    """True if 'name_rel' matches one of the ignore patterns or
    if 'name_rel' refers to a symbolic link"""
    return any([fnmatch(name_rel, pattern) for pattern in setting.ignore]) or \
           (not setting.follow_links and islink(join(dirname, name_rel)))

def generate_tree():
    """generate site tree from files and directories in input directory"""
    logger.info('Ignore files/directories: {}'.format(' '.join(setting.ignore)))
    logger.info('Follow symbolic links: {}'.format(('no','yes')[setting.follow_links]))
    prefix = 'generate:post:page:extension:'
    pagetypes = [key.replace(prefix, '.') for key in event_handler.keys() if key.startswith(prefix)]
    logger.info('Page types: {}'.format(' '.join(pagetypes)))
    folder_queue = {}
    for dirpath, dirnames, filenames in walk(setting.input, topdown=False, followlinks=setting.follow_links):
        dirpath_rel = relpath(dirpath, setting.input)
        folder_name = split(dirpath_rel)[1]
        if dirpath_rel == '.':
            dirpath_rel, folder_name = '', ''
        if ignore_entry(dirpath_rel, setting.input):
            logger.debug(f'Ignore directory {dirpath_rel}')
            continue
        folder = Folder(folder_name, dirpath_rel, None)
        for name in set(dirnames) & set(folder_queue.keys()):
            folder.add(folder_queue[name])
        for name in filenames: # pages and assets; become children of this folder
            filename_rel = name if dirpath_rel == '.' else join(dirpath_rel, name)
            if ignore_entry(filename_rel, setting.input):
                logger.debug(f'Ignore file {filename_rel}')
                continue
            extension = splitext(name)[1]
            this = (Page if extension in pagetypes else Asset)(name, filename_rel, None)
            folder.add(this)
            this.ready()
        folder.ready()
        folder_queue[folder_name] = folder
    # by definition: folder with name = '' is the root of the site tree
    return folder_queue['']



"""
bass.markup
-----
Objects and functions related to markup of text pages.
"""

import re

# available page converters
converter = {}

# Pygments
try:
    import pygments
    have_pygments = True
except:
    have_pygments = False

# Markdown
try:
    import markdown2
    def convert_md2(text):
        extensions = ['tables']
        if have_pygments:
            extensions.append('fenced-code-blocks')
        return markdown2.markdown(text, extras=extensions)
    converter['.mkd'] = convert_md2
    have_markdown  = True
except ImportError:
    have_markdown  = False

if not have_markdown:
    try:
        import markdown
        def convert_mkd(text):
            extras = ['markdown.extensions.tables']
            if have_pygments:
                extras.extend(['markdown.extensions.codehilite', 'markdown.extensions.fenced_code'])
            return markdown.markdown(text, extensions=extras)
        converter['.mkd'] = convert_mkd
        have_markdown = True
    except ImportError:
        have_markdown = False

# RestructuredText
try:
    import docutils.core
    from docutils.writers.html4css1 import Writer
    def convert_rst(text):
        return docutils.core.publish_parts(text, writer=Writer())['body']
    converter['.rst'] = convert_rst
    have_rest = True
except ImportError:
    have_rest = False

# Textile
try:
    import textile
    def convert_txi(text):
        return textile.textile(text)
    converter['.txi'] = convert_txi
    have_textile = True
except ImportError:
    have_textile = False

# HTML
def convert_html(text):
    return text
converter['.html'] = convert_html

# plain text
def convert_txt(text):
    return '<p>' + re.sub(r'\n{2,}', '</p><p>', text) + '</p>'
converter['.txt'] = convert_txt



"""
bass.config
-----
Objects and functions related to configuration.
"""

import argparse
from os import getcwd
from os.path import exists, isfile, join
from . import setting
from .common import read_yaml_file

config_default = dict(follow_links=False, ignore='.?*',
                      host='localhost', port=8080, root_url='/',
                      input='input', output='output', layout='layout')

def read_config():
    """read configuration file, define global settings"""
    config_file = 'config' # assumption: cwd == project directory
    config = config_default.copy()
    if exists(config_file) and isfile(config_file):
        config.update(read_yaml_file(config_file))
    setting.follow_links  = config['follow_links']
    setting.ignore = config['ignore'].split()
    setting.host = config['host']
    setting.port = config['port']
    setting.root_url = config['root_url']
    if config_default['ignore'] not in setting.ignore:
        setting.ignore.append(config_default['ignore'])
    setting.project = getcwd()
    setting.input   = join(setting.project, config['input'])
    setting.layout  = join(setting.project, config['layout'])
    setting.output  = join(setting.project, config['output'])
    if 'extension' in config:
        setting.extension = config['extension']

def parse_cmdline():
    """parse command line, return parsed argument list"""
    parser = argparse.ArgumentParser()
    parser.add_argument('-b', '--build',   help='build',   action='store_true', default=False)
    parser.add_argument('-v', '--version', help='version', action='store_true', default=False)
    parser.add_argument('-c', '--create',  help='create',  action='store_true', default=False)
    parser.add_argument('-d', '--debug',   help='debug',   action='store_true', default=False)
    parser.add_argument('-s', '--server',  help='server',  action='store_true')
    return parser.parse_args()



"""
bass.event
-----
Objects and functions related to events and event handlers.
"""

import re, sys
from datetime import datetime, date
from os.path import join, splitext, getctime, basename
from . import setting
from .markup import converter
from .common import logger

event_handler = {}

def combine(h1, h2):
    def _h12(node):
        h1(node)
        h2(node)
    return _h12

def add_handler(event, handler):
    """add handler for event"""
    if callable(handler):
        if event in event_handler:
            logger.debug(f'Event handler for {event} extended')
            event_handler[event] = combine(event_handler[event], handler)
        else:
            logger.debug(f'New event handler for {event}')
            event_handler[event] = handler
    else:
        logger.debug(f'Event handler for {event} is not a callable')

def copy_handler(from_event, to_event):
    """copy handler for event 'from_event' to event 'to_event'"""
    if from_event in event_handler:
        logger.debug(f'Event handler for {to_event} copied from {from_event}')
        event_handler[to_event] = event_handler[from_event]
    else:
        logger.debug(f'No event handler for {from_event} - cannot copy')

def remove_handler(event):
    """remove handler for event 'event'"""
    if event in event_handler:
        logger.debug(f'Event handler for {event} removed')
        del event_handler[event]
    else:
        logger.debug(f'No event handler for {event} - cannot remove')

def event(event, node):
    """call handler for event 'event'"""
    if event in event_handler:
        event_handler[event](node)

class Processor:
    def __init__(self, converter=None):
        """construct page processor for given markup converter"""
        self.convert = converter

    def __call__(self, node):
        """convert node.content, node.preview and node.meta, which are set by the node constructor,
           to HTML; set elements of node.meta as attributes of node; set node.url"""
        if self.convert:
            # convert node.content and node.preview
            node.preview = self.convert(node.preview) if node.preview else ''
            node.content = self.convert(node.content)
        # process metadata
        full_path = join(setting.input, node.path)
        node.meta = complete_meta(node.meta, full_path)
        # add metadata as node attributes
        for key, value in node.meta.items():
            setattr(node, key, value)
        page_name = splitext(node.path)[0]
        # node.url is setting.root_url + page_name + HTML extension
        node.url = '{0}{1}.html'.format(setting.root_url, page_name)

def complete_meta(meta, path):
    """complete metadata for node with given path"""
    # title: if this is missing, create one from path
    if 'title' not in meta:
        title = splitext(basename(path))[0]
        meta['title'] = title.replace('-', ' ').replace('_', ' ').capitalize()
    # author: cannot be derived from anything else
    # tags
    if 'tags' in meta:
        if isinstance(meta['tags'], list):
            pass # OK
        elif isinstance(meta['tags'], str):
            meta['tags'] = [tag.strip() for tag in meta['tags'].split()]
    else:
        meta['tags'] = []
    # skin
    if 'skin' not in meta:
        meta['skin'] = 'default'
    # id
    if 'id' not in meta:
        meta['id'] = ''
    # date, time, datetime
    fix_date_time(meta, datetime.fromtimestamp(getctime(path)))
    return meta

def fix_date_time(meta, ctime):
    """fix metadata date, time, datetime based on existing metadata and ctime of node"""
    date_part = meta.get('date')
    time_part = meta.get('time')
    if 'datetime' in meta:
        if date_part is None:
            if isinstance(meta['datetime'], datetime):
                date_part = meta['datetime'].date()
            elif isinstance(meta['datetime'], date):
                date_part = meta['datetime']
        if time_part is None and isinstance(meta['datetime'], datetime):
            time_part = meta['datetime'].time()
    meta['date'] = date_part
    meta['time'] = time_part
    # date & time from datetime override date & time from metadata
    if date_part is not None and time_part is not None:
        meta['datetime'] = datetime(date_part.year, date_part.month,
                date_part.day, time_part.hour, time_part.minute,
                time_part.second, time_part.microsecond, time_part.tzinfo)
    elif date_part is not None:
        meta['datetime'] = datetime(date_part.year, date_part.month, date_part.day)
    else:
        meta['datetime'] = ctime
        meta['date']     = ctime.date()
        meta['time']     = ctime.time()

# define event handlers for the standard page types, depending on which Python packages are installed
if '.mkd' in converter:
    markdown_processor = Processor(converter['.mkd'])
    add_handler('generate:post:page:extension:mkd', markdown_processor)
    copy_handler('generate:post:page:extension:mkd', 'generate:post:page:extension:md')

if '.rst' in converter:
    rest_processor = Processor(converter['.rst'])
    add_handler('generate:post:page:extension:rst', rest_processor)

if '.txi' in converter:
    textile_processor = Processor(converter['.txi'])
    add_handler('generate:post:page:extension:txi', textile_processor)

html_processor = Processor(converter['.html'])
add_handler('generate:post:page:extension:html', html_processor)

text_processor = Processor(converter['.txt'])
add_handler('generate:post:page:extension:txt', text_processor)

# auxiliary functions for extensions

# partition and add_toc can be used in event handlers to produce index pages (with optional pagination).
def partition(lst, size):
    """divide list in list of sub-lists of length <= size"""
    return [lst[offset:offset+size] for offset in range(0, len(lst), size)]

def add_toc(page, nodelist, skin, sep='_', size=10):
    """add a table of contents to the specified node, apply pagination if necessary.

    Parameters:
        - page (Node): node to which table of contents is added
        - nodelist ([Node]): nodes to include in table of contents
        - skin (string|callable):
            - string: name of template to render one node in nodelist
            - callable: callable to render one node in nodelist
    Return value: None
    """
    # determine method for converting node to HTML fragment
    if callable(skin):
        logger.debug('add_toc: skin is callable')
        func = skin
    elif isinstance(skin, str):
        logger.debug('add_toc: skin is template')
        func = setting.template[skin].render
    else:
        logger.critical("Bad parameter 'skin' in function 'add_toc'")
        sys.exit(1)
    # create one HTML fragment per node, then partition the list of fragments in chunks of 'size'
    parts = partition([func(this=node) for node in nodelist], size)
    # create 'prev' and 'next' links
    page.prev, page.next = None, None
    previous = page
    logger.debug(f'add_toc: main page name={page.name} path={page.path}, {len(parts)} parts')
    if len(parts) > 0:
        page.toc = '\n'.join(parts[0])
    else:
        page.toc = ''
    if len(parts) > 1:
        for p, part in enumerate(parts[1:]):
            current = page.copy(sep+str(p+1))
            logger.debug(f'subpage name={current.name} path={current.path}')
            current.toc = '\n'.join(part)
            page.add(current)
            previous.next = current
            current.prev = previous
            previous = current
        previous.next = None # last subpage

# resolve_idref is an event handler for resolving idref notation in href attributes.
idref_regex = re.compile(r"href=(['\"])idref:\s*(\w+?)\1")

def resolve_idref(node):
    """replace href='idref:FOO' with href='BAR', where BAR is the URL of the page with id=FOO"""
    def idref_replace(mo):
        catch = node.root().pages(idref=mo.group(2), deep=True)
        return "href={0}{1}{0}".format(mo.group(1), catch[0].url if catch else '#')

    node.preview = idref_regex.sub(idref_replace, node.preview)
    node.content = idref_regex.sub(idref_replace, node.content)



"""
bass.tree
-----
Objects and functions related to the site tree.
"""

import shutil, sys
from copy import copy as shallow_copy
from os import makedirs
from os.path import join, splitext
from . import setting
from .common import read_file, read_yaml_string, write_file, logger
from .event import event

# available assets transformers
transformer = {
    '*': shutil.copy
}

# More complicated examples:
# - browserify vue/dist/vue.runtime.common.js > bundle.js
# - browserify axios/index.js > bundle.js
# - coffee -t -p index.coffee|cs > index.js

def add_transformer(extension, transform):
    """add transformer for asset type (extension)"""
    if callable(transform):
        if extension in transformer:
            # do not use 'combine', since transformations are not by definition commutable
            logger.debug(f'Attempt to redefine transformer for {extension}')
        else:
            logger.debug(f'New transformer for {extension}')
            transformer[extension] = transform
    else:
        logger.debug(f'Transformer for {extension} is not a callable')

# node classes
class Node:
    """Node is the base class for Folder, Page and Asset
       Instance variables:
           - kind: type of node
           - id: identifier of node (usually unique within the tree)
           - name: name of node (last part of path)
           - path: path of node
           - parent: parent node
           - children: list of child nodes

       Instance methods:
           - render: abstract method
           - root: find root of tree
    """
    def __init__(self, name, path, parent=None):
        """construct Node with given name, path and parent"""
        self.kind = 'Node'
        self.id = ''
        self.name = name
        self.path = path
        self.parent = parent
        self.children = []
        self.tags = []

    def ready(self):
        """abstract ready method"""
        pass

    def render(self):
        """abstract render method"""
        pass

    def add(self, node):
        """add child node"""
        node.parent = self
        self.children.append(node)

    def root(self): # follow parent chain until you get 'None'
        """find root of tree in which this node lives"""
        this = self
        while this.parent is not None:
            this = this.parent
        return this


class Folder(Node):
    def __init__(self, name, path, parent):
        """create new Folder node"""
        super().__init__(name, path, parent)
        self.kind = 'Folder'

    def asset(self, name):
        """return asset node with given name in this folder"""
        matches = [child for child in self.children
                   if child.name == name and child.kind == 'Asset']
        return matches[0] if matches else None

    def assets(self):
        """return all asset nodes in this folder"""
        return [child.name for child in self.children if child.kind == 'Asset']

    def folder(self, name):
        """return folder node with given name in this folder"""
        matches = [child for child in self.children
                   if child.name == name and child.kind == 'Folder']
        return matches[0] if matches else None

    def folders(self):
        """return all folder nodes in this folder"""
        return [child for child in self.children if child.kind == 'Folder']

    def page(self, name):
        """return page node with given name in this folder"""
        matches = [child for child in self.children
                   if child.name == name and child.kind == 'Page']
        return matches[0] if matches else None

    def _pages(self, deep=False):
        result = [node for node in self.children if node.kind == 'Page']
        if deep:
            for f in self.folders():
                result.extend(f._pages(deep=True))
        return result

    def pages(self, tag=None, idref=None, deep=False, key='name'):
        """return page nodes with given tag in this folder"""
        result = self._pages(deep=deep)
        if tag:
            result = [node for node in result if tag in node.tags]
        elif idref:
            result = [node for node in result if idref == node.id]
        return sorted(result, key=lambda page: getattr(page, key))

    def ready(self):
        """folder is ready: send event(s)"""
        event('generate:post:root' if self.name == '' else 'generate:post:folder:path:'+self.path, self)

    def render(self):
        """render folder"""
        event('render:pre:root' if self.name == '' else 'render:pre:folder:path:'+self.path, self)
        if self.name != '': # root -> output directory, which already exists
            # rendering a folder means: create sub-directory 'self.path' in output directory
            dirpath = join(setting.output, setting.root_url[1:], self.path)
            logger.debug(f"Creating directory {dirpath}")
            makedirs(dirpath)
        for node in self.children:
            node.render()
        event('render:post:root' if self.name == '' else 'render:post:folder:path:'+self.path, self)


class Page(Node):
    def __init__(self, name, path, parent):
        """create new Page node; set content, preview and meta attributes"""
        super().__init__(name, path, parent)
        self.kind = 'Page'
        # attributes 'skin' and 'url' are derived from metadata by the page processor
        self.skin = ''
        self.url = ''
        full_path = join(setting.input, path)
        self.meta, self.preview, self.content = read_page(full_path)

    def copy(self, sep='_'):
        """create copy of page node, with its own name, path and URL, and empty children list"""
        newpage = shallow_copy(self)
        newpage.children = []
        (page_name, suffix) = splitext(newpage.path)
        newpage.name += sep
        newpage.path = page_name + sep + suffix
        newpage.url = f'{setting.root_url}{page_name}{sep}.html'
        return newpage

    def ready(self):
        """page is ready: send event(s)"""
        suffix = splitext(self.path)[1][1:]
        event('generate:post:page:path:'+self.path, self)
        event('generate:post:page:extension:'+suffix, self)

    def render(self):
        """render Page node"""
        event('render:pre:page:any', self)
        event('render:pre:page:name:' + self.name, self)
        if self.id: event('render:pre:page:id:'+self.id, self)
        for tag in self.tags: event('render:pre:page:tag:'+tag, self)
        # 'skin' attribute should be set by page processor
        if self.skin in setting.template:
            template = setting.template[self.skin]
        else:
            logger.critical(f"Template '{self.skin}' for page {self.path} not available.")
            sys.exit(1)
        filepath = join(setting.output, self.url[1:])
        logger.debug('Writing page {}'.format(filepath))
        write_file(template.render(this=self), filepath)
        for node in self.children: # (dynamically created) sub-pages
            node.render()
        event('render:post:page:any', self)
        event('render:post:page:name:' + self.name, self)
        if self.id: event('render:post:page:id:'+self.id, self)
        for tag in self.tags: event('render:post:page:tag:'+tag, self)


class Asset(Node):
    def __init__(self, name, path, parent):
        """create new Asset node"""
        super().__init__(name, path, parent)
        self.kind = 'Asset'
        self.url = setting.root_url + self.path

    def ready(self):
        """asset is ready: send event(s)"""
        suffix = splitext(self.path)[1][1:]
        event('generate:post:asset:name:'+self.name, self)
        event('generate:post:asset:extension:'+suffix, self)

    def render(self):
        """render Asset node"""
        suffix = splitext(self.path)[1][1:]
        event('render:pre:asset:name:'+self.name, self)
        event('render:pre:asset:extension:'+suffix, self)
        output_path = join(setting.output, setting.root_url[1:], self.path)
        logger.debug(f'Writing asset {output_path}')
        transform = transformer[suffix] if suffix in transformer else transformer['*']
        transform(join(setting.input, self.path), output_path)
        event('render:post:asset:name:'+self.name, self)
        event('render:post:asset:extension:'+suffix, self)

def read_page(path):
    """read page from file and return triple (meta, preview, content)"""
    text = read_file(path)
    parts = text.split('\n---\n')
    if len(parts) == 1: # no metadata, just content
        return {}, '', parts[0]
    elif len(parts) == 2: # metadata, content
        meta = read_yaml_string(parts[0])
        return meta, '', parts[1]
    else: # len(parts) > 2 -> metadata, preview, content
        meta = read_yaml_string(parts[0])
        return meta, parts[1], '\n'.join(parts[1:])



"""
bass.server
-----
Simple web server for development and test purposes.
"""

from os import chdir, walk
from os.path import join, getmtime
from datetime import datetime
from .site import rebuild_site
from . import setting
from .common import logger

try:
    from webob import Request, Response
    from webob.static import DirectoryApp

    class Monitor:
        """class for generating WSGI middleware handler"""
        def __init__(self, app, checklist, callback):
            self.wrapped = app
            self.timestamp = datetime.now()
            self.checklist = checklist
            self.callback = callback

        def changed(self):
            """return True if any file in one of the directories in self.checklist has changed
               since self.timestamp, otherwise False"""
            for entry in self.checklist:
                for (dirpath, _, filenames) in walk(entry):
                    for f in filenames:
                        path = join(dirpath, f)
                        if datetime.fromtimestamp(getmtime(path)) > self.timestamp:
                            logger.debug(f'File {path} has changed')
                            return True
            return False

        def __call__(self, environ, start_response):
            """this __call__ method turns an instance into a WSGI middleware handler"""
            request = Request(environ)
            # check for modifications every time a page (not an asset) is requested
            if request.path.endswith('.html') and self.changed():
                logger.debug('Rebuilding site')
                self.timestamp = datetime.now()
                self.callback()
            response = request.get_response(self.wrapped)
            dt = datetime.strftime(datetime.now(), "%d/%b/%Y %H:%M:%S")
            logger.info('{} - - [{}] "{} {}" {} {}'.\
                        format(request.server_name, dt, request.method, request.path_info,
                               response.status, response.content_length))
            return response(environ, start_response)

    try:
        from waitress import serve
    except ImportError:
        from wsgiref.simple_server import make_server
        def serve(app, host, port):
            """serve: WSGI server with same interface as waitress.serve"""
            server = make_server(host, port, app)
            server.serve_forever()

    def http_server(host, port):
        """http_server: WSGI-based web server with same interface as in standard library"""
        static = DirectoryApp(setting.output, index_page=None)
        wrapped = Monitor(static, checklist=[setting.input, setting.layout], callback=rebuild_site)
        logger.info(f'Starting HTTP server (+watcher) on port {port}')
        serve(wrapped, host=host, port=port)

except ImportError:
    from http.server import HTTPServer, SimpleHTTPRequestHandler
    def http_server(host, port):
        """http_server: basic web server based on standard library"""
        chdir(setting.output)
        httpd = HTTPServer((host, port), SimpleHTTPRequestHandler)
        logger.info(f'Starting HTTP server on port {port}')
        httpd.serve_forever()



#!/usr/bin/env python3

from setuptools import setup

setup(
    name='bass',
    version='1.3.0',
    author='Nico Poppelier',
    author_email='n.poppelier@xs4all.nl',
    license='MIT',
    url='http://schier7.home.xs4all.nl/bass',
    description='Static site generator',
    long_description=
        "Bass is a static website generator. It turns a collection of "
        "content pages, assets (PNG, CSS, JS etc.) into a static website. "
        "You run it on your local computer, and it generates a directory "
        "of files that you can upload to your web server, or serve directly. ",
    download_url="http://schier7.home.xs4all.nl/bass/download",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    install_requires=[
        'chameleon>=3.0',
        'markdown2>=2.3.0',
        'pyyaml>=5.1'
    ],
    packages=['bass'],
    scripts=['script/bass'],
)



"""
bass.setting
-----
Global settings, shared by all modules.
"""

extension    = None
follow_links = None
host         = None
ignore       = None
input        = None
layout       = None
output       = None
port         = None
project      = None
root_url     = None
template     = None
version      = '1.0.0'



"""
Bass
-----
Bass is a tool for building static web sites.
Wok, Wintersmith, Pelican and StrangeCase served as sources of inspiration.
Markdown, RestructuredText and Textile are used for lightweight page markup.
Chameleon is used for templating, but other template engines can be added.

Functions:
    - parse_cmdline: parse command line, return parsed argument list
    - build_site: build new site from content and layout directories
    - create_project: create new project with default configuration
    - http_server: simple HTTP server, for development and testing
    - add_handler: add event handler
    - copy_handler: copy event handler
    - remove_handler: remove event handler
    - add_template_type: add template type (template factory + extension)
    - copy_template_type: copy template type (existing template factory + new extension)
    - logger: logging object
    - resolve_idref: event handler for resolving idref notation
    - add_toc: helper function for creating event handler
"""

from .common  import logger
from .config  import parse_cmdline
from .event   import add_toc, add_handler, copy_handler, remove_handler, resolve_idref
from .layout  import add_template_type, copy_template_type
from .server  import http_server
from .site    import build_site, create_project
from .tree    import add_transformer
